/*
 * Copyright (C) 1995-2002 Neil Peter Charley.
 * Portions Copyright (C) 1996 James William Hawtin.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * dns_server.c
 *
 *  An implementation of a program to do reverse DNS lookups in the
 * background.
 *
 */

char                     dnserverc_rscid[] = "$Revision: 1.8 $";

#if defined(__STRICT_ANSI__)
 #include <features.h>
 #define __USE_POSIX
 #define __USE_BSD
 #define __USE_MISC 
#endif /* __STRICT_ANSI__ */

#include "config.h"

#include <unistd.h>
#if defined(HAVE_FILIOH)
#include <sys/filio.h>
#endif /*
          HAVE_FILIOH 
        */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "dns.h"
#if defined(LOTHLORIEN)
#include "mem.h"
#else /*
         !LOTHLORIEN 
       */
#define ZERO(c, s) memset(c, 0, s)
#endif /*
          LOTHLORIEN 
        */

/*
   Extern Function Prototypes 
 */

/*
   Extern Variables 
 */

#ifndef errno
extern int               errno;
#endif

#if !defined(OSF)
#if defined(BSD)
  /*
     Missing BSD PROTOTYPES 
   */
extern int               setrlimit(int, struct rlimit *);
extern int               getrlimit(int, struct rlimit *);
extern int               setitimer(int, struct itimerval *, struct itimerval *);
extern int               fclose();
extern int               select();
extern time_t            time(time_t *);
extern int               connect();
extern int               shutdown(int, int);
extern int               socket(int, int, int);
extern int               strftime(char *, int, char *, struct tm *);
extern void              perror(char *);
extern int               fflush();
extern int               ioctl();
extern int               fprintf();
extern void              bzero(void *, int);
extern void              bcopy(char *, char *, int);
extern int               vfprintf(FILE * stream, char *format,...);

#endif /*
          BSD 
        */
#else /*
         OSF 
       */
extern int               ioctl(int d, unsigned long request, void *arg);

#endif /*
          !OSF 
        */

#if !defined(MALLOC)
#define MALLOC(c) malloc(c)
#endif /*
          MALLOC 
        */

#if !defined(FREE)
#define FREE(c) free(c)
#endif /*
          FREE 
        */


/*
   Local Function Prototypes 
 */

void                     catch_sigchld(SIGTYPE);
void                     catch_sigterm(SIGTYPE);
void                     check_requests(void);
void                     do_dns_child(int i);
void                     do_request(int i);
void                     log(char *,...);
void                     process_request(int slot);
void                     process_result(int slot);
void                     queue_request(dns_identifier id,
				       struct in_addr *inadd);
void                     restart_child(void);
void                     shutdown_parent(void);
void                     start_child(int slot);
void                     take_off_queue(int freeslot);
void                     write_request(dns_request * dns_req);

/*
   Local Variables 
 */

#define BUFFER_SIZE 1024
char                     scratch[512];
char                     req_buf[BUFFER_SIZE];
dns_child                dns_in_progress[MAX_DNS_IN_PROGRESS];
dns_request             *first_inq = NULL;
dns_request             *last_inq = NULL;
int                      debug = 0;
int                      status = 0;
int                      beats_per_second;
int                      req_size;
pid_t                    parent_pid;

#define DNS_STATUS_RUNNING 1
#define DNS_STATUS_SHUTDOWN 2
#define DNS_STATUS_CHILD_DIED 4

int                      main(int argc, char *argv[])
{
	int                      i;
	struct rlimit            rlp;
#if defined(USE_SIGACTION)
	struct sigaction         sigact;
#endif
	struct sockaddr_in       sadd;
	FILE                    *fp;
	fd_set                   fd_r;
	struct timeval           timeout;

	getrlimit(RLIMIT_NOFILE, &rlp);
	rlp.rlim_cur = rlp.rlim_max;
	setrlimit(RLIMIT_NOFILE, &rlp);

	/* First things first, let the client know we're alive */
	write(1, DNS_SERVER_CONNECT_MSG, strlen(DNS_SERVER_CONNECT_MSG));

	/*
	 * Need this to keep in sync with the client side
	 * <start><dns_id><s_addr> 
	 */
	req_size = sizeof(char) + sizeof(dns_identifier)
	                        + sizeof(sadd.sin_addr.s_addr);

	if (NULL == (fp = fopen(DNS_SERVER_PID_PATH, "w")))
	{
		fprintf(stderr, "Couldn't open '%s' to store Process ID\n",
			DNS_SERVER_PID_PATH);
		perror("dns_server");
		exit(0);
	}
	parent_pid = getpid();
	fprintf(fp, "%d", parent_pid);
	fclose(fp);

	/* Set up signal handling */
#if defined(USE_SIGACTION)
 #if defined(USE_SIGEMPTYSET)
	sigemptyset(&(sigact.sa_mask));
	sigact.sa_handler = (void *)0;
  #else
   #if !defined(__GLIBC__) || (__GLIBC__ < 2)
	sigact.sa_mask = 0;
   #endif
   #if defined(__linux__)
	sigact.sa_restorer = (void *)0;
   #endif
  #endif /* USE_SIGEMPTYSET */
	sigact.sa_handler = catch_sigterm;
	sigaction(SIGTERM, &sigact, (struct sigaction *)0);
	sigact.sa_handler = SIG_IGN;
	sigaction(SIGPIPE, &sigact, (struct sigaction *)0);
	sigact.sa_handler = catch_sigchld;
	sigaction(SIGCHLD, &sigact, (struct sigaction *)0);
#else /* !USE_SIGACTION */
	signal(SIGTERM, catch_sigterm);
	signal(SIGPIPE, SIG_IGN);
	signal(SIGCHLD, catch_sigchld);
#endif /* USE_SIGACTION */

	ZERO(&dns_in_progress[0], MAX_DNS_IN_PROGRESS * sizeof(dns_request));
	/* Pre-fork the children */
	for (i = 0; i < MAX_DNS_IN_PROGRESS; i++)
	{
		start_child(i);
	}
	log("dns", "DNS Parent: Forked %d children\n", MAX_DNS_IN_PROGRESS);
	/* Now enter the main loop */
	status = DNS_STATUS_RUNNING;
	while (status != DNS_STATUS_SHUTDOWN)
	{
		if (1 == getppid())
		{
			/*
			 * If our parent is now PID 1 (init) the talker must
			 * have died without killing us, so we have no
			 * business still being here.
			 * Time to die... 
			 */
			shutdown_parent();
			exit(0);
		}
		FD_ZERO(&fd_r);
		/* STDIN, i.e. from our parent */
		FD_SET(0, &fd_r);
		for (i = 0; i < MAX_DNS_IN_PROGRESS; i++)
		{
			if (dns_in_progress[i].request)
			{
				FD_SET(dns_in_progress[i].DNS_PARENT_READ, &fd_r);
			} else
			{
				/* Free slot, so lets try to fill it */
				take_off_queue(i);
			}
		}
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
		i = select(FD_SETSIZE, &fd_r, 0, 0, &timeout);
		switch (i)
		{
		case -1:
#if defined(DEBUG_DNS_TOO)
			fprintf(stderr, "dns: select failed\n");
#endif /* DEBUG_DNS_TOO */
			break;
		case 0:
			break;
		default:
			if (FD_ISSET(0, &fd_r))
			{
				check_requests();
			}
			for (i = 0; i < MAX_DNS_IN_PROGRESS; i++)
			{
				if (FD_ISSET(dns_in_progress[i].DNS_PARENT_READ,
					     &fd_r))
				{
					process_result(i);
				}
			}
		}

		if (status & DNS_STATUS_CHILD_DIED)
		{
			restart_child();
			status &= ~DNS_STATUS_CHILD_DIED;
		}
	}

	shutdown_parent();
	return 0;
}

/*
 * Catch a SIGTERM, this means to shutdown 
 */

void                     catch_sigterm(SIGTYPE)
{
	status = DNS_STATUS_SHUTDOWN;
}

/*
 * Catch a SIGCHLD, restart if needs be 
 */

void                     catch_sigchld(SIGTYPE)
{
#if !defined(USE_SIGACTION)
	signal(SIGCHLD, catch_sigchld);
#endif /* USE_SIGACTION */
	if (status == DNS_STATUS_SHUTDOWN)
	{
		return;
	}
	status |= DNS_STATUS_CHILD_DIED;
}


/*
 * Check for requests from the client 
 */

void                     check_requests(void)
{
	char                     msgbuf[129];
	int                      toread;
	int                      bread;
	static int               bufpos = 0;
	int                      i;

	ioctl(0, FIONREAD, &toread);
	if (toread <= 0)
	{
		return;
	}
	if (toread > 128)
	{
		toread = 128;
	}
	bread = read(0, msgbuf, toread);
	for (i = 0; i < bread;)
	{
		req_buf[bufpos++] = msgbuf[i++];
		if (bufpos == req_size)
		{
			process_request(bufpos);
			bufpos = 0;
		}
	}
}

void                     process_request(int toread)
{
	int                      i;
	struct in_addr           inadd;
	dns_identifier           dns_id;

	for (i = 0; i < toread;)
	{
		switch (req_buf[i++])
		{
		case DNS_CLIENT_SEND_REQUEST:
			memcpy(&dns_id, &req_buf[i], sizeof(dns_id));
			i += sizeof(dns_id);
			memcpy(&(inadd.s_addr), &req_buf[i],
			       sizeof(inadd.s_addr));
			i += sizeof(inadd.s_addr);

#if defined(DEBUG_DNS_TOO)
			fprintf(stderr, "Server: Id      = %d\n"
				"        Address = %08X\n"
				"Bytes needed %d, Bytes to recieve %d\n",
				dns_id,
				(int)ntohl(inadd.s_addr),
				i, toread);
			fflush(stderr);
#endif /* DEBUG_DNS_TOO */
			queue_request(dns_id, &inadd);
			break;
		case DNS_CLIENT_CANCEL_REQUEST:
			memcpy(&dns_id, &req_buf[i], sizeof(dns_id));
			i += sizeof(dns_id);
			break;
		default:
#if defined(DEBUG_DNS_TOO)
			fprintf(stderr, "DNS_Server: Unknown message from client:"
				" '%d'\n",
				req_buf[i]);
#endif /* DEBUG_DNS_TOO */
			break;
		}
	}
}


/*
 * Queue up a request from the client and/or send it off 
 */

void                     queue_request(dns_identifier id, struct in_addr *inadd)
{
	dns_request             *new_dns_req;

	new_dns_req = (dns_request *) MALLOC(sizeof(dns_request));
	new_dns_req->dns_id = id;
	new_dns_req->next = NULL;
	memcpy(&(new_dns_req->ip_addr), inadd,
	       sizeof(struct in_addr));

#if defined(DEBUG_DNS)
	log("dns", "DNS Parent: Queueing %s\n",
	    inet_ntoa(new_dns_req->ip_addr));
#endif /* DEBUG_DNS */
	if (!first_inq)
	{
		first_inq = new_dns_req;
	} else if (!last_inq)
	{
		last_inq = new_dns_req;
		first_inq->next = last_inq;
	} else
	{
		last_inq->next = new_dns_req;
		last_inq = new_dns_req;
	}
#if defined(DEBUG_DNS_TOO)
	log("dns", "DNS Parent: Queued %s; first = %08X, last = %08X\n",
	    inet_ntoa(new_dns_req->ip_addr), first_inq, last_inq);
#endif /* DEBUG_DNS_TOO */
}

void                     take_off_queue(int freeslot)
{
	if (first_inq)
	{
		dns_in_progress[freeslot].request = first_inq;
		first_inq = first_inq->next;
#if defined(DEBUG_DNS_TOO)
		log("dns", "DNS Parent: Sending '%s' to child %d\n",
		    inet_ntoa(dns_in_progress[freeslot].request->ip_addr),
		    freeslot);
#endif /* DEBUG_DNS_TOO */
		do_request(freeslot);
	}
}


/*
 * Send request to DNS child process 
 */

void                     do_request(int i)
{
	int                      bwritten = 0;

#if defined(DEBUG_DNS_TOO)
	log("dns", "Parent: do_request %s\n",
	    inet_ntoa(dns_in_progress[i].request->ip_addr));
#endif /* DEBUG_DNS_TOO */

	while (bwritten < sizeof(struct in_addr))
	{
		bwritten += write(dns_in_progress[i].DNS_PARENT_WRITE,
			    (char *)&(dns_in_progress[i].request->ip_addr),
				  sizeof(struct in_addr) - bwritten);

	}
	if (bwritten == -1)
	{
		perror("Parent write");
	}
#if defined(DEBUG_DNS_TOO)
	log("dns", "Parent: do_request %s (wrote %d bytes, on fd %d) DONE\n",
	    inet_ntoa(dns_in_progress[i].request->ip_addr), bwritten,
	    dns_in_progress[i].DNS_PARENT_WRITE);
#endif /* DEBUG_DNS_TOO */
}


void                     process_result(int slot)
{
	int                      bread;
	char                    *s, *t;

	s = scratch;
	*s++ = DNS_SERVER_SEND_REPLY;
	memcpy(s, &(dns_in_progress[slot].request->dns_id),
	       sizeof(dns_identifier));
	s += sizeof(dns_identifier);
#if defined(DEBUG_DNS_TOO)
	fprintf(stderr, "DNS Server: Processing result...\n");
	fflush(stderr);
#endif /* DEBUG_DNS_TOO */

	t = s;
	/* Loop reading the reply in until we get the \n */
	while (*(t - 1) != '\n' && (t - scratch) < 300)
	{
		bread = read(dns_in_progress[slot].DNS_PARENT_READ, t, 256);
		t += bread;
	}

	*t = '\0';
	/* Now actually send the reply to the client */
	write(1, scratch, (t - scratch));
#if defined(DEBUG_DNS)
	log("dns_ids", "SERVER: slot %d, dns_id %d, reply '%s'\n",
	    slot, dns_in_progress[slot].request->dns_id, s);
#endif /* DEBUG_DNS */
#if defined(DEBUG_DNS_TOO)
	fprintf(stderr, "    Sending reply '%s' of %d bytes\n",
		s, (t - scratch));
	fflush(stderr);
#endif /* DEBUG_DNS_TOO */
	FREE(dns_in_progress[slot].request);
	dns_in_progress[slot].request = NULL;
}


/*
 * Do a loop, acting as a child doing actual lookups 
 */

void                     do_dns_child(int i)
{
	char                     buf[sizeof(struct in_addr)];
	struct in_addr           inadd;
	int                      toread;
	int                      bread;
	int                      bufpos = 0;
	struct hostent          *host;
	char                     reply[258];	 /* 256 from rfc822 + \n\0 */
	struct timeval           timeout;
	fd_set                   fds;
	int                      sel_ret;
	int                      attempts;
	int											 ip;

	while (1)
	{
		FD_ZERO(&fds);
		FD_SET(dns_in_progress[i].DNS_CHILD_READ, &fds);
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;
		sel_ret = select(FD_SETSIZE, &fds, 0, 0, &timeout);
		if (getppid() != parent_pid)
		{
			/*
			 * Parent must have died, 'cos it
			 * is no longer our parent. Time to die...
			 */
			exit(1);
		}
		switch (sel_ret)
		{
		case -1:	/* Error */
			if (errno != EINTR)
			{
				log("dns_debug",
				    "Child %d: select error'd\n",
				    i);
			}
			break;
		case 0:	/* Nothing ready */
			if (timeout.tv_usec != 0 &&
			    timeout.tv_sec != 0)
			{
				/*
				log("dns_debug", "Child %d: nothing ready, not timed out\n", i);
				*/
			}
			continue;
		}
		/* Wait for request */
		ioctl(dns_in_progress[i].DNS_CHILD_READ, FIONREAD, &toread);
		if (toread <= 0)
		{
			if (errno != EINTR)
			{
				log("dns_debug", "Child %d: toread <= 0\n", i);
			}
			continue;
		}
		if ((toread + bufpos) > sizeof(struct in_addr))
		{
			toread = sizeof(struct in_addr) - bufpos;
		}
		bread = read(dns_in_progress[i].DNS_CHILD_READ, buf,
			     sizeof(struct in_addr) - bufpos);

		bufpos += bread;
		if (bufpos != sizeof(struct in_addr))
		{
			log("dns", "Child: not read sizeof(struct in_addr)\n");
			continue;
		}
		memcpy((void *)&inadd, (void *)buf, sizeof(struct in_addr));

#if defined DEBUG_DNS_TOO
		log("dns", "DNS child %d: Doing lookup on %s\n",
		    i, inet_ntoa(inadd));
#endif /* DEBUG_DNS_TOO */
		attempts = 0;
		do
		{
			attempts++;
			/* Do lookup */
			host = gethostbyaddr((char *)&inadd, sizeof(inadd),
					     AF_INET);
			/* Put together return */
			if (!host)
			{
				switch (h_errno)
				{
				case NETDB_INTERNAL:
					log("dns", "NETDB_INTERNAL: %s\n",
					    inet_ntoa(inadd));
					attempts = DNS_LOOKUP_ATTEMPTS;
					break;
				case HOST_NOT_FOUND:
					attempts = DNS_LOOKUP_ATTEMPTS;
					break;
				case TRY_AGAIN:
					log("dns", "TRY_AGAIN(%d): %s\n",
					    attempts, inet_ntoa(inadd));
					/* Retry a number of times */
					break;
				case NO_RECOVERY:
					log("dns", "NO_RECOVERY: %s\n",
					    inet_ntoa(inadd));
					attempts = DNS_LOOKUP_ATTEMPTS;
					break;
				case NO_DATA:	/* Shouldn't be possible for PTR's */
					log("dns", "NO_DATA: %s\n",
					    inet_ntoa(inadd));
					attempts = DNS_LOOKUP_ATTEMPTS;
					break;
				default:
					log("dns", "Unknown error: %s\n",
					    inet_ntoa(inadd));
					attempts = DNS_LOOKUP_ATTEMPTS;
				}
			} else
			{
				sprintf(reply, "%s", host->h_name);
			}
		}
		while (attempts < DNS_LOOKUP_ATTEMPTS && !host);
		if (host)
		{
			/* Now, as we got a hostname back, see if that resolves to the same IP */
			attempts = 0;
			do
			{
				attempts++;
				/* Do lookup */
				host = gethostbyname(reply);
				/* Put together return */
				if (!host)
				{
					switch (h_errno)
					{
					case NETDB_INTERNAL:
						log("dns", "NETDB_INTERNAL: %s\n", reply);
						attempts = DNS_LOOKUP_ATTEMPTS;
						break;
					case HOST_NOT_FOUND:
						attempts = DNS_LOOKUP_ATTEMPTS;
						break;
					case TRY_AGAIN:
						log("dns", "TRY_AGAIN(%d): %s\n", attempts, reply);
						/* Retry a number of times */
						break;
					case NO_RECOVERY:
						log("dns", "NO_RECOVERY: %s\n", reply);
						attempts = DNS_LOOKUP_ATTEMPTS;
						break;
					case NO_DATA:	/* Shouldn't be possible for PTR's */
						log("dns", "NO_DATA: %s\n", reply);
						attempts = DNS_LOOKUP_ATTEMPTS;
						break;
					default:
						log("dns", "Unknown error: %s\n", reply);
						attempts = DNS_LOOKUP_ATTEMPTS;
					}
				} else
				{
					/* Check if any of the returned IPs for the name match what
					 * we were given to start with
					 */
					for (ip = 0 ; host->h_addr_list[ip] ; ip++)
					{
						/* if (!memcmp(&inadd, (struct in_addr *)*host->h_addr_list[ip], sizeof(struct in_addr)))*/
						if (inadd.s_addr == ((struct in_addr *) host->h_addr_list[ip])->s_addr)
						{
						/* Got a match */
							break;
						}
					}
					if (!host->h_addr_list[ip])
					{
						/* We didn't find a match, so back to the IP */
						sprintf(reply, "%s", inet_ntoa(inadd));
					}
				}
			}
			while (attempts < DNS_LOOKUP_ATTEMPTS && !host);
		}
		if (!host)
		{
			sprintf(reply, "%s", inet_ntoa(inadd));
		}
#if defined DEBUG_DNS_TOO
		log("dns", "DNS child %d: Lookup on %s returned '%s'\n",
		    i, inet_ntoa(inadd), reply);
#endif /* DEBUG_DNS_TOO */
		/* Send response */
		strcat(reply, "\n");
		write(dns_in_progress[i].DNS_CHILD_WRITE, reply, strlen(reply));
		bufpos = 0;
	}
}


/*
 * Make sure all the children are shutdown 
 */

void                     shutdown_parent(void)
{
	int                      i;

	for (i = 0; i < MAX_DNS_IN_PROGRESS; i++)
	{
		kill(dns_in_progress[i].pid, SIGCONT);
	}
	for (i = 0; i < MAX_DNS_IN_PROGRESS; i++)
	{
		kill(dns_in_progress[i].pid, SIGTERM);
	}
	for (i = 0; i < MAX_DNS_IN_PROGRESS; i++)
	{
		waitpid(dns_in_progress[i].pid, NULL, WNOHANG);
	}
}


/*
 * Start up a child, in the given slot 
 */

void                     start_child(int slot)
{
	if (-1 == pipe(dns_in_progress[slot].to_child_fds))
	{
		switch (errno)
		{
		case EMFILE:
			log("boot", "dns_server_main: Too many fd's"
			    " in use, trying to fork child %d\n",
			    slot);
			exit(1);
		case ENFILE:
			log("boot", "dns_server_main: Too many fd's"
			    " in use in system.\n");
			exit(1);
		case EFAULT:
			log("boot", "dns_server_main: to_child_fds"
			    " invalid!\n");
			exit(1);
		}
	}
	if (-1 == pipe(dns_in_progress[slot].from_child_fds))
	{
		switch (errno)
		{
		case EMFILE:
			log("boot", "dns_server_main: Too many fd's"
			    " in use, trying to fork child %d\n",
			    slot);
			exit(1);
		case ENFILE:
			log("boot", "dns_server_main: Too many fd's"
			    " in use in system.\n");
			exit(1);
		case EFAULT:
			log("boot", "dns_server_main: from_child_fds"
			    " invalid!\n");
			exit(1);
		}
	}
	dns_in_progress[slot].pid = fork();
	switch (dns_in_progress[slot].pid)
	{
	case -1:	/* Error */
		/* Aardvarks */
		log("boot", "DNS Server: Couldn't fork child %d\n",
		    slot);
		exit(1);
		break;
	case 0:	/* Child */
		close(dns_in_progress[slot].DNS_PARENT_WRITE);
		close(dns_in_progress[slot].DNS_PARENT_READ);
		dns_in_progress[slot].DNS_PARENT_WRITE =
			dns_in_progress[slot].DNS_PARENT_READ = -1;
		do_dns_child(slot);
		/* NOT REACHED */
		break;
	default:	/* Parent */
		close(dns_in_progress[slot].DNS_CHILD_READ);
		close(dns_in_progress[slot].DNS_CHILD_WRITE);
		dns_in_progress[slot].DNS_CHILD_READ =
			dns_in_progress[slot].DNS_CHILD_WRITE = -1;
	}
}


/*
 * Find, and attempt to restart a child that has died 
 */

void                     restart_child(void)
{
	pid_t                    pid;
	int                      pstatus;
	int                      i;

	pid = waitpid(-1, &pstatus, WNOHANG);
	if (pid > 0)
	{
		for (i = 0; i < MAX_DNS_IN_PROGRESS; i++)
		{
			if (dns_in_progress[i].pid != pid)
			{
				continue;
			}
			close(dns_in_progress[i].DNS_PARENT_WRITE);
			close(dns_in_progress[i].DNS_PARENT_READ);
			start_child(i);
			if (dns_in_progress[i].request)
			{
				/* Argh, was doing something */
				do_request(i);
			} else
			{

			}
		}
	} else
	{
		log("dns", "DNS Server: SIGCHLD, waitpid() == %d\n",
		    pid);
	}
}


/*
 * log using stdarg variable length arguments 
 */

void                     log(char *str,...)
{
	va_list                  args;
	FILE                    *fp;
	char                    *fmt;
	struct tm               *tim;
	char                     the_time[256];
	char                     fname[21];
	time_t                   current_time;

	va_start(args, str);

	current_time = time(0);
	tim = localtime(&current_time);
	strftime((char *)&the_time, 255, "%H:%M:%S %Z - %d/%m/%Y - ",
		 tim);
	sprintf(fname, "logs/%s.log", str);

	fp = fopen(fname, "a");
	if (!fp)
	{
		fprintf(stderr, "Eek! Can't open file to log: %s\n", fname);
		return;
	}
	fprintf(fp, "%s", the_time);
	fmt = va_arg(args, char *);
	vfprintf(fp, fmt, args);
	va_end(args);
	fclose(fp);
}
